/*
 * acedecoder.c: ac3 interface for DVD player
 *
 * Copyright (C) 2000 Matjaz Thaler
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at matjaz.thaler@rd.iskraemeco.si 
 *
 * the project's page is at http://www2.arnes.si/~mthale1/index.html
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include <errno.h>
#include <unistd.h>

#include "dvdplayer.h"

int data_read_ac3(unsigned char *data, int datasize);
int get_ac3_frame(unsigned char *data, int *sync);
void wait_for_data();

int ac3_dec(void){
  FILE *ac3pipe;
  unsigned char buff[4096+2];
  int sync=0;
  int frameSize=1024;
  int delay;
  float Delay=data_buffer->ac3sync*frameSize;

  int i=1;

  ac3pipe=popen("ac3dec", "w");
  if(ac3pipe==NULL){
    fprintf (stderr, "can't run ac3dec\n");
    data_buffer->exit_child=1;
    return(1);
  }
  while(1){
    i++;
    Delay=((data_buffer->input_pointer_ac3-data_buffer->output_pointer_ac3)%AC3_BUFFER_SIZE)*0.002+0.998*Delay;
    delay=(int)(Delay+0.5);
    if(data_buffer->ac3stat==AC3_START){
      //printf("start\n");
      data_buffer->output_pointer_ac3=data_buffer->input_pointer_ac3;
      sync=0;
      wait_for_data();
      Delay=data_buffer->ac3sync*frameSize;
      data_buffer->ac3stat=AC3_PLAY;
    }
    else if(data_buffer->ac3stat==AC3_STOP){
      data_buffer->output_pointer_ac3=data_buffer->input_pointer_ac3;
      sync=0;
      //printf("stop\n");
      wait_for_data();
      continue;
    }
    else if(delay >= (data_buffer->ac3sync+3)*frameSize && data_buffer->ac3sync != -1 && data_buffer->video != 0 
	    && data_buffer->ac3stat==AC3_PLAY && !(i%128)){
      //printf("skip frame\n");
      if(delay >= (data_buffer->ac3sync+8)*frameSize){
	get_ac3_frame(buff, &sync);
	get_ac3_frame(buff, &sync);
	get_ac3_frame(buff, &sync);
	Delay-=3*frameSize;
	//printf("skip 3 frames\n");
      }
      get_ac3_frame(buff, &sync);
      Delay-=frameSize;
    }
    else if(delay <= (data_buffer->ac3sync-4)*frameSize && data_buffer->ac3sync != -1 && data_buffer->video != 0 
	    && data_buffer->ac3stat==AC3_PLAY && !(i%4)){
      //printf("wait for new frame\n");
      wait_for_data();
    }
    
    frameSize=get_ac3_frame(buff, &sync);
    if(frameSize){
      fwrite(buff, frameSize, 1, ac3pipe);
      // if(!(i%256))
	//printf("frame_size: %4.4d delay: %2.2d %2.2d\n", frameSize, (int)(delay/frameSize), data_buffer->ac3sync);
    }
    
  }
  pclose(ac3pipe);
  return 0;
}

int data_read_ac3(unsigned char *data, int datasize){
  int datasize2;

  datasize2=0;
  while(datasize){
    if(data_buffer->output_pointer_ac3 != data_buffer->input_pointer_ac3){ // data in the buffer
      data[datasize2++]=data_buffer->ac3_data[data_buffer->output_pointer_ac3];
      datasize--;
      data_buffer->output_pointer_ac3=
        (data_buffer->output_pointer_ac3+1)%AC3_BUFFER_SIZE;
    }
    else{                        // buffer empty
      if(sem_ctl(semId, 1, 2))   
	exit(-1);
      if(sem_op(semId, 1, 0))    // wait for data
	exit(-1);
    }
  }
  if(sem_ctl(semId, 1, 1))       // wake up main if sleeps
    exit(-1);
  if(data_buffer->exit_child){
    exit(-1);
  }
  return(datasize2);
}

int get_ac3_frame(unsigned char *data, int *syn){

  int i, j=10;

  while(j){
    if(!(*syn)){
      while(1){
	data_read_ac3(&data[0], 1);
	if(data[0]==0x0b){
	  data_read_ac3(&data[1], 1);
	  if(data[1]==0x77){
	    *syn=1;
	    //printf("find ac3 sync\n");
	    break;
	  }
	}
      }
    }
    else{
      data[0]=0x0b;
      data[1]=0x77;
    }
    i=0;
    while(i<16){
      data_read_ac3(&data[256*i+2], 256);
      i++;
      if(data[256*i]==0x0b && data[256*i+1]==0x77)
	return i*256;
    }
    //printf("can't find ac3 sync %d\n", j);
    *syn=0;
    j--;
  }
  *syn=0;
  fprintf(stderr, "can't find ac3 sync\n");
  return 0;
}

void wait_for_data(){

  if(sem_ctl(semId, 1, 2))   
    exit(-1);
  if(sem_op(semId, 1, 0))
    exit(-1);
}
